local tec = TeamEquipCompare;
local libSerialize = LibStub:GetLibrary("AceSerializer-3.0");

-- methods that should be part of lua
function tec_keys(t, sorted)
	local result = {};

	for k,v in pairs(t) do result[#result + 1] = k end
	if (sorted) then sort(result); end

	return result;
end

function tec_vals(t)
	local result = {};

	for k,v in pairs(t) do result[#result + 1] = v end

	return result;
end

function tec_containskey(t, key)
	if (type(t) == 'table') then
		return (t[key] ~= nil);
	else
		return false;
	end
end

function tec_round(n)
	local f = floor(n);
	if ( n >= (f + 0.5)) then
		return ceil(n);
	else
		return f;
	end
end
-- end lua-esque methods

function TEC_RequestPartyData()
	if (IsInGroup()) then
		tec:notify("Requesting data from party...");
		tec:SendCommMessage(tec.commPrefix, "SYNC", "PARTY");
	else
		tec:notify("Unable to request team member data -- not in a party");
	end
end

function TEC_RequestUnitData(name)
	tec:notify(format("Requesting data from '%s'...", name));
	tec:SendCommMessage(tec.commPrefix, "REQUEST", "WHISPER", name);
end

function TEC_SendPlayerData(sendTo)
	local data = TEC_GetPlayerData();

	if (sendTo == "PARTY") then
		tec:notify("Sending data to party...");
		tec:SendCommMessage(tec.commPrefix, TEC_PackageData(data), "PARTY");
	else
		tec:notify(format("Sending data to '%s'...", sendTo));
		tec:SendCommMessage(tec.commPrefix, TEC_PackageData(data), "WHISPER", sendTo);
	end
end

function TEC_GetPlayerData()
	tec:debug("Building player data...");
	
	local ilvlTotal, ilvlEquipped = GetAverageItemLevel();
	local t = {
		name = GetUnitName("player", false),
		realm = GetRealmName(),
		class = UnitClass("player"),
		spec = TEC_GetSpecName(),
		level = UnitLevel("player"),
		ilevel = {
			total = ilvlTotal,
			equipped = ilvlEquipped
		},
		money = GetMoney(),
		gear = {},
	};
	
	for ii = 1, 18 do
		link = GetInventoryItemLink(t.name, ii);
		if (link) then
			t.gear[ii] = GetInventoryItemLink(t.name, ii);
		end
	end

	tec:debug(format("Player inspection completed (%d items)", #t.gear));

	return t;
end

function TEC_SendLevelUpdate(level)
	tec:debug("Sending level update to party...");
	tec:SendCommMessage(tec.commPrefix, "LEVEL="..level, "PARTY");
end

function TEC_SendMoneyUpdate(money)
	tec:debug("Sending money update to party...");
	tec:SendCommMessage(tec.commPrefix, "MONEY="..money, "PARTY");
end

function TEC_SendSpecUpdate(spec)
	tec:debug("Sending spec update to party...");
	tec:SendCommMessage(tec.commPrefix, "SPEC="..spec, "PARTY");
end

function TEC_PrintTeam()
	if (not IsInGroup()) then
		tec:Print("No team member data found -- not in group");
	else
		local names =  tec_keys(tec.party, true);
		if (#names == 0) then
			tec:Print("No team member data found -- type '/tec sync' to force a data sync");
		else
			local totalMoney = 0;
			for index, name in pairs(names) do
				local data = tec.party[name];
				if (data) and (data.name) then
					totalMoney = totalMoney + (data.money or 0);

					if (tec.db.global.trackMoney) then
						tec:Print(format("|c%s%s|r [%d]: spec=%s, ilevel=%d, money=%s", TEC_ClassColor_Lookup[data.class], name, data.level, data.spec, tec_round(data.ilevel.equipped), GetCoinTextureString(data.money)));
					else
						tec:Print(format("|c%s%s|r [%d]: spec=%s, ilevel=%d", TEC_ClassColor_Lookup[data.class], name, data.level, data.spec, tec_round(data.ilevel.equipped)));
					end
				else
					tec:Print(format("|cff70b8ff%s|r: no data ('/tec sync' will force a data sync)", name));
				end
			end

			if (tec.db.global.trackMoney) then
				tec:Print(format("Team total money: %s", GetCoinTextureString(totalMoney)));
			end
		end
	end
end

function TEC_CheckTeamMembers()
	local count = GetNumGroupMembers();
	local unitsToAdd = {};
	local unitsToRemove = {};
	local playerName = GetUnitName("player", false);
	
	-- first check if local player is included
	if (not tec_containskey(tec.party, playerName)) then
		unitsToAdd[#unitsToAdd + 1] = playerName;
	end

	-- next check for each party member
	for ii=1,count-1 do
		local name = GetUnitName("party"..tostring(ii));
		if (not tec_containskey(tec.party, name)) then
			unitsToAdd[#unitsToAdd + 1] = name;
		end
	end

	-- find orphans
	for unit, _ in pairs(tec.party) do
		if (not UnitInParty(unit)) then
			unitsToRemove[#unitsToRemove + 1] = unit;
		end
	end

	return unitsToAdd, unitsToRemove;
end

function TEC_UpdateSavedDB(db, default)
	-- loop through all keys
	for k, v in pairs (default) do
		-- update missing value
		if (not tec_containskey(db, k)) then
			tec:debug(format("key '%s' is missing, assigning default value...", k));
			db[k] = v;
		else
			-- recursively walk tables
			if (type(v) == 'table') then
				return TEC_UpdateSavedDB(db[k], v);
			end
		end
	end

	return db;
end

function TEC_GetSpecName()
	local spec = "None";

	local specid = GetSpecialization();
	if (specid) then
		spec = select(2, GetSpecializationInfo(specid));
	end

	return spec;
end

function TEC_PackageData(data)
	local serialized = libSerialize:Serialize(data);
	return serialized;
end

function TEC_UnpackageData(packaged)
	local success, data = libSerialize:Deserialize(packaged);
	if (success) then
		return data;
	else
		return "";
	end
end